/*
	vobj2tray.c

	original code: SASANO Takayoshi
	--- public domain, no warranty.
*/

#include <btron/btron.h>
#include <btron/hmi.h>
#include <tcode.h>
#include <errcode.h>
#include <bstdio.h>
#include <bstdlib.h>
#include <tstring.h>
#include <tcode.h>

/* $B%(%s%G%#%"%sJQ49$,I,MW$K$J$C$?;~$N$?$a$N!"3HD%MQ(B */
#define	XeH(x)	(x)
#define	XeW(x)	(x)

/* $B:GBg%H%l!<%l%3!<%I?t(B */
#define	TRAYREC_MAX	8192	// SYSCONF:VObjMax$B!_(B2$BDxEY$NCM$H$9$k(B

/* $B%H%l!<%l%3!<%I3JG<%(%j%"(B */
LOCAL	TRAYREC	Tray[TRAYREC_MAX];
LOCAL	W	TrayPos = 0;

/* $B2~9T%3!<%IDI2C$NM-L5(B */
LOCAL	W	AddNL;

/* $B%G%P%C%0MQ(B */
#ifdef	DEBUG
#define	DP(x)	printf x
#else
#define	DP(x)	/* */
#endif
#define	P(x)	printf x

/* $B%G%P%C%0;Y1g(B */
#ifdef	DEBUG
typedef	struct {
	UB	*str;
	W	id;
} SegName;

LOCAL	const SegName	NameTbl[] = {
	{"TR_VOBJ",	0x01},
	{"TR_TEXT",	0x02},
	{"TR_FIG",	0x03},
	{"TS_INFO",	0xE0},
	{"TS_TEXT",	0xE1},
	{"TS_TEXTEND",	0xE2},
	{"TS_FIG",	0xE3},
	{"TS_FIGEND",	0xE4},
	{"TS_IMAGE",	0xE5},
	{"TS_VOBJ",	0xE6},
	{"TS_DFUSEN",	0xE7},
	{"TS_FFUSEN",	0xE8},
	{"TS_SFUSEN",	0xE9},
	{"TS_TPAGE",	0xA0},
	{"TS_TRULER",	0xA1},
	{"TS_TFONT",	0xA2},
	{"TS_TCHAR",	0xA3},
	{"TS_TATTR",	0xA4},
	{"TS_TSTYLE",	0xA5},
	{"TS_TVAR",	0xAD},
	{"TS_TMEMO",	0xAE},
	{"TS_TAPPL",	0xAF},
	{"TS_FPRIM",	0xB0},
	{"TS_FDEF",	0xB1},
	{"TS_FGRP",	0xB2},
	{"TS_FMAC",	0xB3},
	{"TS_FATTR",	0xB4},
	{"TS_FPAGE",	0xB5},
	{"TS_FMEMO",	0xBE},
	{"TS_FAPPL",	0xBF},
	{"unknown",	~0},
};
LOCAL	UB	*find_segname(W id)
{
	W	i;

	for (i = 0; i < sizeof(NameTbl) / sizeof(SegName) - 1; i++) {
		if (id == NameTbl[i].id) break;
	}

	return NameTbl[i].str;
}
#endif

/* $BJ8;zNs(B($BC1?t(B)$B"*J8;zNs(B($BJ#?t(B)$B$NHf3S(B */
LOCAL	W	tc_strcmpX(TC *s1, TC *s2[])
{
	for (; *s2 != NULL; s2++) {
		if (!tc_strcmp(s1, *s2)) return 0;
	}

	return 1;
}

/* $BJ8>O3+;O%;%0%a%s%H$r%H%l!<%l%3!<%I$KDI2C(B */
LOCAL	W	add_textseg(void)
{
	LOCAL	const UH	ts[] = {
		XeH(0), XeH(0), XeH(0), XeH(0),	// $BI=<(NN0h(B
		XeH(0), XeH(0), XeH(0), XeH(0),	// $BIA2hNN0h(B
		XeH(0x0000),			// $B?eJ?%f%K%C%H(B
		XeH(0x0000),			// $B?bD>%f%K%C%H(B
		XeH(0x0021),			// $B%G%U%)!<%k%H8@8l!JF|K\8l!K(B
		XeH(0x0000),			// $BGX7J%Q%?!<%s(BID$B!JF)L@!K(B
	};

	Tray[TrayPos].id = TS_TEXT;
	Tray[TrayPos].len = sizeof(ts);
	Tray[TrayPos].dt = (UB *)ts;

	TrayPos++;

	return (TrayPos >= TRAYREC_MAX);
}

/* $B2~9T%3!<%I$r%H%l!<%l%3!<%I$KDI2C(B */
LOCAL	W	add_newline(void)
{
	LOCAL	const UH	nl = XeH(TC_NL);

	Tray[TrayPos].id = TR_TEXT;
	Tray[TrayPos].len = sizeof(nl);
	Tray[TrayPos].dt = (UB *)&nl;

	TrayPos++;

	return (TrayPos >= TRAYREC_MAX);
}

/* $B2>?H%l%3!<%I$r%H%l!<%l%3!<%I$KDI2C(B */
LOCAL	W	add_vobjrec(W fd, VOBJSEG *vseg, W len, TC *target[])
{
	W	err, asize;
	TR_VOBJREC	*v;
	TC	vname[L_FNM + 1];

	asize = sizeof(VLINK) + len;

	/* $B%j%s%/%l%3!<%I!&2>?H%;%0%a%s%H3JG<NN0h$N3NJ](B */
	v = malloc(asize);
	if (v == NULL) {
		P(("add_vobjrec: malloc NULL\n"));
		err = ER_NOMEM;
		goto fin0;
	}

	/* $B%j%s%/%l%3!<%I$N8!:w$HFI$_=P$7(B */
	err = fnd_rec(fd, F_FWD, (1 << 0), 0, NULL);
	if (err < ER_OK) {
		P(("add_vobjrec: fnd_rec %d\n", err));
		goto fin1;
	}

	err = rea_rec(fd, 0, (UB *)&v->vlnk, sizeof(VLINK), NULL, NULL);
	if (err < ER_OK) {
		P(("add_vobjrec: rea_rec %d\n", err));
		goto fin1;
	}

	/* $B<!$N%l%3!<%I$X0\9T!J%(%i!<%A%'%C%/$O9T$o$J$$!K(B */
	see_rec(fd, 1, 0, NULL);

	/* $B2>?HL>$N%A%'%C%/(B */
	if (target != NULL) {
		err = fil_sts((LINK *)&v->vlnk, vname, NULL, NULL);
		if (err < ER_OK) {
			P(("add_vobjrec: fil_sts %d\n", err));
			goto fin1;
		}

		if (tc_strcmpX(vname, target)) {
			err = ER_OK;
			goto fin1;
		}
	}

	/* $B2>?H%;%0%a%s%H(B */
	memcpy(&v->vseg, vseg, len);

	Tray[TrayPos].id = TR_VOBJ;
	Tray[TrayPos].len = asize;
	Tray[TrayPos].dt = (UB *)v;

	TrayPos++;

	/* $B2~9T%3!<%I$NDI2C(B */
	if (AddNL && TrayPos < TRAYREC_MAX) add_newline();

	err = (TrayPos >= TRAYREC_MAX);
	goto fin0;

 fin1:
	free(v);
 fin0:
	return err;
}

/* $B2>?H%l%3!<%IMQ$K3NJ]$7$?NN0h$r2rJ|(B */
LOCAL	void	free_vobjrec(void)
{
	W	i;

	for (i = 0; i < TrayPos; i++) {
		if (Tray[i].id == TR_VOBJ) free(Tray[i].dt);
	}

	return;
}

/* $B2>?H%;%0%a%s%H$N8!:w(B */
LOCAL	W	scan_vobjseg(W fd, UH *buf, W size, TC *target[])
{
	W	err, i, pos, id, len;

	err = ER_OK;

	for (i = 0; i < size; i++) {
		if ((id = buf[(pos = i)] ) < 0xff00) continue;

		if (buf[pos + 1] == 0xffff) {
			len = buf[pos + 2] | (buf[pos + 3] << 16);
			i += 3;
		} else {
			len = buf[pos + 1];
			i += 1;
		}

#ifdef	DEBUG
		P(("%08x: %04x %08x <%s>\n",
		   pos * 2, id, len, find_segname(id & 0xff)));
#endif

		/* $B2>?H%l%3!<%IH/8+!'%H%l!<$KDI2C(B */
		if (id == (TC_SPEC | TS_VOBJ)) {
			err = add_vobjrec(fd, (VOBJSEG *)&buf[i + 1],
					  len, target);
			if (err) {
				P(("scan_vobjseg: add_vobjrec %d\n", err));
				goto fin0;
			}
		}

		i += len / 2;
	}

 fin0:
#ifdef	DEBUG
	/* $B%G%P%C%0MQ(B */
	for (i = 0; i < TrayPos; i++) {
		DP(("scan_vobjseg: id %08x len %08x dt %08x <%s>\n",
		    Tray[i].id, Tray[i].len, Tray[i].dt,
		    find_segname(Tray[i].id)));
	}
#endif

	return err;
}

/* $B=hM}K\BN(B */
LOCAL	W	vobj2tray(W fd, UH *buf, W size, TC *target[])
{
	W	err;

	/* $B%H%l!<%l%3!<%I3JG<%(%j%"$N=i4|2=(B */
	memset(Tray, 0, sizeof(Tray));

	/* $BJ8>O3+;O%;%0%a%s%H$r5-O?(B */
	add_textseg();

	/* $B2>?H%;%0%a%s%H$r8!:w!&5-O?(B */
	err = scan_vobjseg(fd, buf, size / 2, target);

	/* $B%H%l!<$X=q$-9~$_!J%(%i!<%A%'%C%/$O$7$J$$!K(B */
	if (TrayPos > 1) tpsh_dat(Tray, TrayPos, NULL);

	/* $B2>?H%;%0%a%s%HMQ$K3NJ]$7$?NN0h$N2rJ|(B */
	free_vobjrec();

	return err;
}

/* main */
EXPORT	W	main(W ac, TC *av[])
{
	W	err, fd, size;
	LINK	l;
	VP	buf;

	/* $B0z?t$N%A%'%C%/(B */
	if (ac < 2) {
		P(("usage: %S [filename] [(vobjname)] ...\n", av[0]));
		err = ER_PAR;
		goto fin0;
	}

	/* $B%3%^%s%IL>$N:G8e$NJ8;z$,(BZ$B$N>l9g$O2~9T%3!<%I$rDI2C$7$J$$(B */
	AddNL = (av[0][tc_strlen(av[0]) - 1] != TK_Z);

	/* $BF~NO%U%!%$%k$X$N%j%s%/<h$j=P$7(B */
	err = get_lnk(av[1], &l, F_NORM);
	if (err < ER_OK) {
		P(("main: get_lnk %d\n", err));
		goto fin0;
	}

	/* $BF~NO%U%!%$%k$N%*!<%W%s(B */
	err = opn_fil(&l, F_READ | F_EXCL, NULL);
	if (err < ER_OK) {
		P(("main: opn_fil %d\n", err));
		goto fin0;
	}
	fd = err;

	/* $B<g(BTAD$B%l%3!<%I$X0\F0(B */
	err = fnd_rec(fd, F_ENDTOP, (1 << 1), 0, NULL);
	if (err < ER_OK) {
		P(("main: fnd_rec %d\n", err));
		goto fin1;
	}

	/* $B%l%3!<%I%5%$%:$N<hF@(B */
	err = rea_rec(fd, 0, NULL, 0, &size, NULL);
	if (err < ER_OK) {
		P(("main: rea_rec %d\n", err));
		goto fin1;
	}

	/* $B%a%b%j$N3NJ](B */
	buf = malloc(size);
	if (buf == NULL) {
		P(("main: malloc NULL\n"));
		err = ER_NOMEM;
		goto fin1;
	}

	/* $B%l%3!<%IA4BN$NFI$_9~$_(B */
	err = rea_rec(fd, 0, buf, size, NULL, NULL);
	if (err < ER_OK) {
		P(("main: rea_rec %d\n", err));
		goto fin2;
	}

	/* $B@hF,%l%3!<%I$X0\F0(B */
	err = see_rec(fd, 0, 1, NULL);
	if (err < ER_OK) {
		P(("main: see_rec %d\n", err));
		goto fin2;
	}

	/* $B%H%l!<$X$N=q$-=P$7(B */
	err = vobj2tray(fd, buf, size, (ac < 3) ? NULL : &av[2]);

 fin2:
	free(buf);
 fin1:
	cls_fil(fd);
 fin0:
	return err;
}
